import binascii

from pymacaroons.binders.base_binder import BaseBinder
from pymacaroons.utils import hmac_concat, truncate_or_pad, convert_to_bytes


class HashSignaturesBinder(BaseBinder):

    def bind_signature(self, signature):
        return hmac_concat(
            truncate_or_pad(b'\0'),
            binascii.unhexlify(convert_to_bytes(self.root.signature)),
            binascii.unhexlify(convert_to_bytes(signature))
        )



__author__ = 'Evan Cordell'

__version__ = "0.6.0"
__version_info__ = tuple(__version__.split('.'))
__short_version__ = __version__

from .caveat import Caveat
from .macaroon import Macaroon
from .verifier import Verifier

__all__ = [
    'Macaroon',
    'Caveat',
    'Verifier',
]



from __future__ import unicode_literals

import binascii
import copy
from base64 import standard_b64encode

from libnacl.secret import SecretBox

from pymacaroons.caveat import Caveat
from pymacaroons.binders import HashSignaturesBinder
from pymacaroons.utils import (truncate_or_pad,
                               create_initial_macaroon_signature,
                               generate_derived_key,
                               sign_first_party_caveat,
                               sign_third_party_caveat)


class RawMacaroon(object):
    """
    RawMacaroon uses byte types internally and in its public interface.

    This relegates all string/byte conversion and input validation to
    the to Macaroon proxy class, which is the public interface.

    RawMacaroons should not be constructed directly.

    """

    def __init__(self,
                 location=None,
                 identifier=None,
                 key=None,
                 caveats=None,
                 signature=None):
        self._caveats = []
        if location:
            self._location = location
        if identifier:
            self._identifier = identifier
        if location and identifier and key:
            self._signature = create_initial_macaroon_signature(
                key, identifier
            )
        if signature:
            self._signature = signature

    @property
    def location(self):
        return self._location

    @property
    def identifier(self):
        return self._identifier

    @property
    def signature(self):
        return self._signature

    @property
    def caveats(self):
        return self._caveats

    def copy(self):
        return copy.deepcopy(self)

    def first_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.first_party()]

    def third_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.third_party()]

    # Protects discharge macaroons in the event they are sent to
    # the wrong location by binding to the root macaroon
    def prepare_for_request(self, macaroon):
        return HashSignaturesBinder(self).bind(macaroon)

    # The existing macaroon signature is the key for hashing the
    # caveat being added. This new hash becomes the signature of
    # the macaroon with caveat added.
    def add_first_party_caveat(self, predicate):
        caveat = Caveat(caveatId=predicate)
        self._caveats.append(caveat)
        encode_key = binascii.unhexlify(self.signature)
        self._signature = sign_first_party_caveat(encode_key, predicate)
        return self

    # The third party caveat key is encrypted useing the current signature, and
    # the caveat is added to the list. The existing macaroon signature
    # is the key for hashing the string (verificationId + caveatId).
    # This new hash becomes the signature of the macaroon with caveat added.
    def add_third_party_caveat(self, location, key, key_id, nonce=None):
        derived_key = truncate_or_pad(generate_derived_key(key))
        old_key = truncate_or_pad(binascii.unhexlify(self.signature))
        box = SecretBox(key=old_key)
        encrypted = box.encrypt(derived_key, nonce=nonce)
        verificationKeyId = standard_b64encode(encrypted)
        caveat = Caveat(
            caveatId=key_id,
            location=location,
            verificationKeyId=verificationKeyId
        )
        self._caveats.append(caveat)
        encode_key = binascii.unhexlify(self.signature)
        self._signature = sign_third_party_caveat(
            encode_key,
            caveat._verificationKeyId,
            caveat._caveatId
        )
        return self



from __future__ import unicode_literals
import copy

from pymacaroons.raw_macaroon import RawMacaroon
from pymacaroons.serializers.binary_serializer import BinarySerializer
from pymacaroons.serializers.json_serializer import JsonSerializer
from pymacaroons.exceptions import MacaroonInitException
from pymacaroons.utils import convert_to_bytes, convert_to_string


class Macaroon(object):

    def __init__(self,
                 location=None,
                 identifier=None,
                 key=None):

        if location and identifier and key:
            self._raw_macaroon = RawMacaroon(
                location=convert_to_bytes(location),
                identifier=convert_to_bytes(identifier),
                key=convert_to_bytes(key)
            )
        else:
            raise MacaroonInitException(
                'Must supply all: (location, id, key).'
            )

    @staticmethod
    def from_binary(serialized):
        if serialized:
            return BinarySerializer().deserialize(serialized)
        else:
            raise MacaroonInitException(
                'Must supply binary serialized macaroon.'
            )

    @staticmethod
    def from_json(serialized):
        if serialized:
            return JsonSerializer().deserialize(serialized)
        else:
            raise MacaroonInitException(
                'Must supply json serialized macaroon.'
            )

    @property
    def location(self):
        return convert_to_string(self._raw_macaroon.location)

    @property
    def identifier(self):
        return convert_to_string(self._raw_macaroon.identifier)

    @property
    def signature(self):
        return convert_to_string(self._raw_macaroon.signature)

    @property
    def caveats(self):
        return self._raw_macaroon.caveats

    def copy(self):
        return copy.deepcopy(self)

    # Concatenates location, id, all caveats, and signature,
    # and then base64 encodes them
    def serialize(self):
        return BinarySerializer().serialize(self)

    def serialize_json(self):
        return JsonSerializer().serialize(self)

    def inspect(self):
        combined = 'location {loc}\n'.format(loc=self.location)
        combined += 'identifier {id}\n'.format(id=self.identifier)
        for caveat in self.caveats:
            combined += 'cid {cid}\n'.format(cid=caveat.caveatId)
            if caveat.verificationKeyId and caveat.location:
                combined += 'vid {vid}\n'.format(vid=caveat.verificationKeyId)
                combined += 'cl {cl}\n'.format(cl=caveat.location)
        combined += 'signature {sig}'.format(sig=self.signature)
        return combined

    def first_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.first_party()]

    def third_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.third_party()]

    # Protects discharge macaroons in the event they are sent to
    # the wrong location by binding to the root macaroon
    def prepare_for_request(self, macaroon):
        protected = self.copy()
        protected._raw_macaroon = self._raw_macaroon.prepare_for_request(
            macaroon._raw_macaroon
        )
        return protected

    # The existing macaroon signature is the key for hashing the
    # caveat being added. This new hash becomes the signature of
    # the macaroon with caveat added.
    def add_first_party_caveat(self, predicate):
        self._raw_macaroon.add_first_party_caveat(
            convert_to_bytes(predicate)
        )
        return self

    # The third party caveat key is encrypted useing the current signature, and
    # the caveat is added to the list. The existing macaroon signature
    # is the key for hashing the string (verificationId + caveatId).
    # This new hash becomes the signature of the macaroon with caveat added.
    def add_third_party_caveat(self, location, key, key_id):
        return self._raw_macaroon.add_third_party_caveat(
            convert_to_bytes(location),
            convert_to_bytes(key),
            convert_to_bytes(key_id)
        )




class BaseBinder(object):
    def __init__(self, root):
        self.root = root

    def bind(self, discharge):
        protected = discharge.copy()
        protected._signature = self.bind_signature(discharge._signature)
        return protected

    def bind_signature(self, signature):
        raise NotImplementedError()



import json

from pymacaroons.utils import convert_to_string, convert_to_bytes


class JsonSerializer(object):

    def serialize(self, macaroon):
        serialized = {
            'location': macaroon.location,
            'identifier': macaroon.identifier,
            'caveats': [caveat.to_dict() for caveat in macaroon.caveats],
            'signature': macaroon.signature
        }
        return json.dumps(serialized)

    def deserialize(self, serialized):
        from pymacaroons.raw_macaroon import RawMacaroon
        from pymacaroons.macaroon import Macaroon
        from pymacaroons.caveat import Caveat

        raw_macaroon = RawMacaroon()

        deserialized = json.loads(convert_to_string(serialized))
        raw_macaroon._location = deserialized['location']
        raw_macaroon._identifier = deserialized['identifier']
        for c in deserialized['caveats']:
            caveat = Caveat(
                caveatId=c['cid'],
                verificationKeyId=c['vid'],
                location=c['cl']
            )
            raw_macaroon._caveats.append(caveat)
        raw_macaroon._signature = convert_to_bytes(deserialized['signature'])
        m = Macaroon(location=b'\0', identifier=b'\0', key=b'\0')
        m._raw_macaroon = raw_macaroon
        return m



from __future__ import unicode_literals

import binascii
import struct
from base64 import urlsafe_b64encode, urlsafe_b64decode

from pymacaroons.utils import convert_to_bytes
from pymacaroons.serializers.base_serializer import BaseSerializer


class BinarySerializer(BaseSerializer):

    PACKET_PREFIX_LENGTH = 4

    def serialize(self, macaroon):
        raw_macaroon = macaroon._raw_macaroon
        combined = self._packetize(b'location', raw_macaroon.location)
        combined += self._packetize(b'identifier', raw_macaroon.identifier)

        for caveat in raw_macaroon.caveats:
            combined += self._packetize(b'cid', caveat._caveatId)

            if caveat._verificationKeyId and caveat._location:
                combined += self._packetize(b'vid', caveat._verificationKeyId)
                combined += self._packetize(b'cl', caveat._location)

        combined += self._packetize(
            b'signature',
            binascii.unhexlify(raw_macaroon._signature)
        )
        return urlsafe_b64encode(combined).decode('ascii')

    def deserialize(self, serialized):
        from pymacaroons.raw_macaroon import RawMacaroon
        from pymacaroons.macaroon import Macaroon
        from pymacaroons.caveat import Caveat
        from pymacaroons.exceptions import MacaroonDeserializationException

        raw_macaroon = RawMacaroon()
        decoded = urlsafe_b64decode(convert_to_bytes(
            serialized + "=" * (-len(serialized) % 4)
        ))

        index = 0

        while index < len(decoded):
            packet_length = int(
                struct.unpack(
                    b"4s",
                    decoded[index:index + self.PACKET_PREFIX_LENGTH]
                )[0],
                16
            )
            packet = decoded[
                index + self.PACKET_PREFIX_LENGTH:index + packet_length
            ]

            key, value = self._depacketize(packet)

            if key == b'location':
                raw_macaroon._location = value
            elif key == b'identifier':
                raw_macaroon._identifier = value
            elif key == b'cid':
                raw_macaroon.caveats.append(Caveat(caveatId=value))
            elif key == b'vid':
                raw_macaroon.caveats[-1].verificationKeyId = value
            elif key == b'cl':
                raw_macaroon.caveats[-1].location = value
            elif key == b'signature':
                raw_macaroon._signature = convert_to_bytes(
                    binascii.hexlify(value)
                )
            else:
                raise MacaroonDeserializationException(
                    'Key {key} not valid key for this format. Value: '.format(
                        key=key, value=value
                    )
                )

            index = index + packet_length

        m = Macaroon(location=b'\0', identifier=b'\0', key=b'\0')
        m._raw_macaroon = raw_macaroon
        return m

    def _packetize(self, key, data):
        # The 2 covers the space and the newline
        packet_size = self.PACKET_PREFIX_LENGTH + 2 + len(key) + len(data)
        # Ignore the first two chars, 0x
        packet_size_hex = hex(packet_size)[2:]
        header = packet_size_hex.zfill(4).encode('ascii')
        packet_content = key + b' ' + data + b'\n'
        packet = struct.pack(
            convert_to_bytes("4s%ds" % len(packet_content)),
            header,
            packet_content
        )
        return packet

    def _depacketize(self, packet):
        key = packet.split(b' ')[0]
        value = packet[len(key) + 1:-1]
        return (key, value)
